#!/home/gchossie/anaconda3/envs/freight_us/bin/python

#SBATCH --cpus-per-task=16
#SBATCH --mem=60G
#SBATCH --time=10:00:00

import pickle
import pandas as pd
import requests
import zipfile
import io
import os
import random
import time
import pycountry
from joblib import Parallel, delayed
import multiprocessing

directory = './raw_data/'
if not os.path.isdir(directory):
    os.mkdir(directory)


def jp_monitor_data(year: int) -> None:
    '''
    Parameters
    ---------

    year (int)
    country (str): 2-letter ISO code (can be obtained with pycountry)

    Fetches all raw monitoring data for all prefectures from Japan's Ministry
    of the Environment. All of 2018 and 2019-01 through 04 are missing.
    '''
    directory = './raw_data/japan/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    if year == 2016 or year == 2017:
        for prov in range(1, 48):
            url = f'http://www.nies.go.jp/igreen/tj/{year}/j{str(prov).zfill(2)}_{year}.zip'
            resp = requests.get(url)
            z = zipfile.ZipFile(io.BytesIO(resp.content))
            z.extractall(f'{directory}{year}/j{str(prov).zfill(2)}_{year}/')
    if year == 2019:
        for mth in range(4, 13):
            for prov in range(1, 48):
                url = f'http://soramame.taiki.go.jp/DownLoad/2019{str(mth).zfill(2)}/2019{str(mth).zfill(2)}_{str(prov).zfill(2)}.zip'
                resp = requests.get(url)
                z = zipfile.ZipFile(io.BytesIO(resp.content))
                z.extractall(f'{directory}{year}/j{str(prov).zfill(2)}_2019/')
    if year == 2020:
        for mth in range(1, 6):
            for prov in range(1, 48):
                url = f'http://soramame.taiki.go.jp/DownLoad/2020{str(mth).zfill(2)}/2020{str(mth).zfill(2)}_{str(prov).zfill(2)}.zip'
                resp = requests.get(url)
                z = zipfile.ZipFile(io.BytesIO(resp.content))
                z.extractall(f'{directory}{year}/j{str(prov).zfill(2)}_2020/')
    print(f'Japan {year}: Success.')
    return None


def eu_monitor_data(year: int, country: str) -> None:
    '''
    Parameters
    ---------

    year (int)
    country (str): 2-letter ISO code (can be obtained with pycountry)
    '''
    directory = './raw_data/europe/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    poll_names = {1: 'SO2', 8: 'NO2', 7: 'O3',
                  5: 'PM10', 6001: 'PM2.5', 10: 'CO'}
    for pollutant_code in poll_names.keys():
        url = f'https://fme.discomap.eea.europa.eu/fmedatastreaming/AirQualityDownload/AQData_Extract.fmw?CountryCode={country}&CityName=&Pollutant={pollutant_code}&Year_from={year}&Year_to={year}&Station=&Samplingpoint=&Source=All&Output=TEXT&UpdateDate=&TimeCoverage=Year'
        resp = requests.get(url)
        with open(f'{directory}list_files_{country}_{year}_{poll_names[pollutant_code]}.csv', 'wb') as f:
            f.write(resp.content)
        try:
            files = pd.read_csv(f'{directory}list_files_{country}_{year}_{poll_names[pollutant_code]}.csv',
                                header=None)
        except:
            continue
        # Initialize with the first file
        try:
            temp = pd.io.parsers.read_csv(files[0].values[0])
        except:
            continue
        for idx in range(1, len(files[0].values)):
            url = files[0].values[idx]
            try:
                temp = pd.concat([temp, pd.io.parsers.read_csv(url)], axis=0)
            except:
                continue

        temp.reset_index().to_feather(f'{directory}{country}_raw_{year}_{poll_names[pollutant_code]}.feather')
    cnt_name = pycountry.countries.get(alpha_2=country).name
    print(f'{cnt_name} {year}: Success.')
    return None


def us_monitor_utility() -> None:
    '''
    Download the most recent files with monitor information
    '''
    directory = './raw_data/us/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    url_monitors = 'https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/today/monitoring_site_locations.dat'
    monloc = pd.read_csv(url_monitors, sep='|', header=None,
                         encoding='ISO-8859-1')
    lst = ['AQSID', 'parametername', 'sitecode', 'sitename', 'status', 'agencyid', 'agencyname',
           'EPAregion', 'latitude', 'longitude', 'elevation', 'GMToffset', 'countrycode', 'MSAcode',
                                    'MSAname', 'statecode', 'statename', 'countycode', 'countyname']
    monloc = monloc.rename(columns={i: lst[i] for i in range(len(lst))})
    # Clean-up
    monloc = monloc.loc[monloc['status'] == 'Active'].reset_index(drop=True)
    monloc = monloc.drop(columns=[19, 20, 21, 22])
    idx = (monloc['longitude'] >= -172.5).values * \
        (monloc['longitude'] <= -36).values
    monloc = monloc.loc[idx]
    monloc.to_csv(directory + 'us_monitoring_site_locations.csv')
    return monloc


def us_monitor_data(year: int) -> None:
    '''
    Parameters
    ---------

    year (int)
    '''
    directory = './raw_data/us/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    startdate = f'{year}0101'
    if year == 2020:
        enddate = f'{year}0531'
    else:
        enddate = f'{year}1231'
    monloc = us_monitor_utility()

    # Initialize final value table
    species = ['O3', 'PM2.5', 'PM10', 'SO2', 'NO2', 'CO']
    vals = dict.fromkeys(species)
    trad = {'O3': 'OZONE-8HR', 'PM2.5': 'PM2.5-24hr',
            'PM10': 'PM10-24hr', 'SO2': 'SO2-24HR'}
    for sp in vals.keys():
        tmpp = monloc.loc[monloc['parametername'] == sp].copy()
        vals[sp] = pd.DataFrame(columns=pd.date_range(start=startdate, end=enddate, freq='D'),
                                index=set(tmpp['AQSID']))

    # Loop through each day
    for dat in pd.date_range(start=startdate, end=enddate, freq='D'):
        # Get files with daily data first
        yy = dat.strftime(format='%Y')
        dd = dat.strftime(format='%Y%m%d')
        if not os.path.exists(directory + f'{dd}.dat'):
            url = f'https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/{yy}/{dd}/daily_data_v2.dat'
            resp = requests.get(url)
            # If data v2 is not available, switch to data v1
            if resp.status_code == 404:
                url = f'https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/{yy}/{dd}/daily_data.dat'
                resp = requests.get(url)
                if resp.status_code == 404:
                    raise ValueError(f'Data could not be acquired on {dat}')
            with open(directory + f'{dd}.dat', 'wb') as fil:
                fil.write(resp.content)

        raw = pd.read_csv(directory + f'{dd}.dat',
                          sep='|', header=None, encoding='latin1')
        raw = raw.rename(columns={0: 'Date', 1: 'AQSID', 2: 'Location', 3: 'Metric', 4: 'Unit',
                                  5: 'Value', 6: 'Averaging period', 7: 'Data source', 8: 'AQI value',
                                  9: 'AQI category', 10: 'Latitude', 11: 'Longitude', 12: 'Full AQSID'})
        # Add the new daily data to the output dict for species with daily values
        for sp in species[:-2]:
            tmp = raw.loc[raw['Metric'] == trad[sp]].set_index(
                'AQSID', drop=True).copy()
            for idx in tmp.index:
                if idx in vals[sp].index:
                    vals[sp].loc[idx, dat] = tmp.loc[idx, 'Value']

        # Get files that only have hourly data
        lst_cols = ['Valid date', 'validtime', 'AQSID', 'sitename',
                    'GMToffset', 'parametername', 'reportingunits', 'value', 'datasource']
        temp_daily = {'NO2': pd.DataFrame(
            columns=lst_cols), 'CO': pd.DataFrame(columns=lst_cols)}
        for ddat in pd.date_range(pd.to_datetime(f'{dd}00', format='%Y%m%d%H'),
                                  pd.to_datetime(f'{dd}23', format='%Y%m%d%H'), freq='H'):
            h = ddat.strftime(format='%Y%m%d%H')
            if not os.path.exists(directory + f'{dd}_{h[-2:]}00.dat'):
                url = f'https://s3-us-west-1.amazonaws.com//files.airnowtech.org/airnow/{yy}/{dd}/HourlyData_{h}.dat'
                time.sleep(random.random())
                resp = requests.get(url)
                if resp.status_code != 200:
                    raise ValueError(f'URL not found: {url}')
                with open(directory + f'{dd}_{h[-2:]}00.dat', 'wb') as fil:
                    fil.write(resp.content)

            raw = pd.read_csv(directory + 
                f'{dd}_{h[-2:]}00.dat', sep='|', header=None, encoding='latin1')
            raw = raw.rename(columns={i: lst_cols[i]
                                      for i in range(len(lst_cols))})
            temp_daily['NO2'] = pd.concat(
                [temp_daily['NO2'], raw.loc[raw['parametername'] == 'NO2']], axis=0)
            temp_daily['CO'] = pd.concat(
                [temp_daily['CO'], raw.loc[raw['parametername'] == 'CO']], axis=0)

        # Process and add to table
        for sp in species[-2:]:
            tt = temp_daily[sp].sort_values(by='validtime').groupby('AQSID')[
                'value'].mean()
            for idx in tt.index:
                try:
                    vals[sp].loc[idx, dat] = tt.loc[idx]
                except:
                    pass
    tdy = datetime.datetime.today().strftime(format='%b_%d')
    with open(directory + f'{year}_us_raw.pkl', 'wb') as f:
        pickle.dump(vals, f)
    print(f'US {year}: Success.')
    return None


def ch_monitor_data(year: int) -> None:
    '''
    Parameters
    ---------

    year (int)
    '''
    directory = './raw_data/china/'
    if not os.path.isdir(directory):
        os.mkdir(directory)
    startdate = f'{year}0101'
    if year == 2020:
        enddate = f'{year}0531'
    else:
        enddate = f'{year}1231'
    dates = pd.date_range(start=startdate, end=enddate, freq='D')
    temp = pd.DataFrame([])
    for dat in dates:
        strdat = dat.strftime(format='%Y%m%d')
        url = f'http://beijingair.sinaapp.com/data/china/sites/{strdat}/csv'
        temp = pd.concat([temp, pd.read_csv(url)], axis=0)
        time.sleep(random.random())
    temp.to_csv(directory + f'{year}_raw_china.csv')
    print(f'China {year}: Success.')
    return None


def kr_monitor_data(year: int) -> None:
    '''
    Parameters
    ---------

    year (int)

    This takes time: the 2020 data is 56MB, previous years are ~200 MB
    '''
    directory = './raw_data/korea/'
    if not os.path.isdir(directory):
        os.mkdir(directory)

    url = f'https://www.airkorea.or.kr/jfile/readDownloadFile.do?fileId=1657eebba9237&fileSeq={str(year)[-2:]}'
    headers = {'Host': 'www.airkorea.or.kr',
               'Connection': 'keep-alive',
               'Upgrade-Insecure-Requests': '1',
               'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36',
               'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
               'Sec-Fetch-Site': 'same-origin',
               'Sec-Fetch-Mode': 'navigate',
               'Sec-Fetch-User': '?1',
               'Sec-Fetch-Dest': 'document',
               'Referer': 'https://www.airkorea.or.kr/web/last_amb_hour_data?pMENU_NO=123',
               'Accept-Encoding': 'gzip, deflate, br',
               'Accept-Language': 'en-US,en;q=0.9',
               'Cookie': 'WMONID=D2jZJJF0hpI; tm_x=197923.160699; tm_y=451713.403762; station=111121; JSESSIONID=8iFAdhxmUSZF7ULqZuuRXdbjtRM2RNRlvbAsdxEt39N9GP6v44NVvyQmY1u9TxrA.airwas1_servlet_newairkorea4'}
    resp = requests.get(url, stream=True, headers=headers)
    with open(directory + str(year) + '.zip', 'wb') as fd:
        for chunk in resp.iter_content(chunk_size=128):
            fd.write(chunk)
    z = zipfile.ZipFile(open(directory + str(year) + '.zip', 'rb'))
    z.extractall(f'{directory}{year}/')
    print(f'South Korea {year}: Success.')
    return None


def kr_monitor_data_recent(year=2020):
    '''
    Parameters
    ---------

    year (int)

    This scrapes monthly data for the most recent months from the website
    '''
    directory = './raw_data/korea/'
    spe = {'PM2.5': '10008', 'O3': '10003', 'NO2': '10006'}
    areacodes = ['02', '051', '053', '032', '062', '042', '031', '033', '043', '041', '063',
                 '061', '044', '054', '055', '064']
    ymds = ['2020-04-30', '2020-05-31']
    tdates = ['2020-04-01', '2020-05-01']
    url = 'https://www.airkorea.or.kr/web/sidoAirInfo/sidoAirInfoDay30'
    for sp in list(spe.keys()):
        for area in areacodes:
            for i in range(len(ymds)):
                params = {'itemCode': spe[sp], 'ymd': ymds[i],
                          'areaCode': area, 'tDate': tdates[i], 'monthDay': '31'}
                resp = requests.get(url, params)
                if resp.content == 404:
                    raise ValueError(f'URL {url} is invalid')
                new = pd.read_html(resp.content)[0]
                # with open(f'./data/kr_raw/2020/{sp}_{area}.pkl', 'rb') as f:
                #    old = pickle.load(f)
                temp = new.drop(index=[0]).T
                temp = temp.rename(
                    columns={i: temp.loc['지점', i] for i in range(1, len(list(temp))+1)})
                temp = temp.T.set_index('지점', drop=True).T
                #new = pd.concat([old,temp],axis=0)
                with open(directory + f'{ymds[i][:4]}/{ymds[i][:4]}_{ymds[i][5:7]}_{sp}_{area}.pkl', 'wb') as f:
                    pickle.dump(temp, f)
    print(f'South Korea {year}: Success.')
    return None


def get_oxford_strigency():
    raw = pd.read_csv('https://github.com/OxCGRT/covid-policy-tracker/raw/master/data/OxCGRT_latest.csv',
                      parse_dates=['Date'])
    raw = raw[['CountryName', 'CountryCode', 'Date', 'C1_School closing',
               'C2_Workplace closing', 'C3_Cancel public events',
               'C4_Restrictions on gatherings', 'C5_Close public transport',
               'C6_Stay at home requirements', 'C7_Restrictions on internal movement',
               'C8_International travel controls', 'H1_Public information campaigns',
               'StringencyIndex']]
    raw = raw.fillna(method='pad')
    raw = pd.pivot_table(raw, values='StringencyIndex', index='Date',
                         columns='CountryName')
    raw = raw.loc[raw.index <= pd.to_datetime('20200520')]
    with open('./raw_data/shutdown_index.pkl', 'wb') as f:
        pickle.dump(raw, f)
    return None

def wrapper(region, year):
    if region == 'United States':
        us_monitor_data(year)
    elif region == 'Europe':
        eu_monitor_data(year)
    elif region == 'China':
        ch_monitor_data(year)
    elif region == 'Japan':
        jp_monitor_data(year)
    elif region == 'South Korea':
        kr_monitor_data(year)
        kr_monitor_data_recent()

if __name__ == '__main__':
    to_process = []
    for region in ['United States', 'Europe', 'China', 'South Korea', 'Japan']:
        for year in range(2016,2021):
            to_process.append((region, year))
    num_cores = multiprocessing.cpu_count()
    results = Parallel(n_jobs=num_cores)(delayed(wrapper)(region, year) for region, year in to_process)
    get_oxford_strigency()

